DrDamage = AceLibrary("AceAddon-2.0"):new("AceConsole-2.0", "AceEvent-2.0", "AceDB-2.0", "AceHook-2.1")
--local L = AceLibrary("AceLocale-2.2"):new("DrDamage");
local SEA = AceLibrary("SpecialEvents-Aura-2.0")
local GT = AceLibrary("Gratuity-2.0")

local defaults = {
	SpellDamage = 0,
	Healing = 0,
	ShadowDamage = 0,
	FireDamage = 0,
	NatureDamage = 0,
	FrostDamage = 0,
	ArcaneDamage = 0,
	HolyDamage = 0,
	ItemCrit = 0,
	ItemHit = 0,
	DisplayType = "Avg",
	ManualDmg = false,
	ManualPlus = false,
	FontSize = 8,
	FontPosition = 0,
}

DrDamage:RegisterDB("DrDamageDB");
DrDamage:RegisterDefaults("profile", defaults )

function DrDamage:OnInitialize()	
end

local changesAB = {
    "ACTIONBAR_PAGE_CHANGED",
    "ACTIONBAR_SLOT_CHANGED",
    "UNIT_INVENTORY_CHANGED",
    "PLAYER_TARGET_CHANGED",
    "UNIT_AURA",
}

local changesDMG = {
    "CHARACTER_POINTS_CHANGED",
    "LEARNED_SPELL_IN_TAB",
    "PLAYER_ALIVE",
}

function DrDamage:DrOnEvent()

	for i = 1, #changesAB do
    	
	    	if changesAB[i] and changesAB[i] == event then
	    	    	
	    		if event == "UNIT_AURA" then
	    		
	    			if arg1 == "target" or arg1 =="player" then
	    			
	    				if not self:IsEventScheduled("UpdatingAB") then
	    					self:ScheduleEvent("UpdatingAB", self.UpdateAB, 1.5, self)
	    				end
	    			end
	    		else
	    			if not self:IsEventScheduled("UpdatingAB") then
	    				self:ScheduleEvent("UpdatingAB", self.UpdateAB, 0.5, self)
	    			end
	    		end
	    	end
	end
	
	for i = 1, #changesDMG do
    	
    		if changesDMG[i] and changesDMG[i] == event then
    			DrDamage:DmgUpdateTalents()
    			DrDamage:UpdateAB()
    		end
    	end
end

function DrDamage:Compatible( check )

	if check == "Class" then
		localizedClass, englishClass = UnitClass("player");
		
		if englishClass == "WARLOCK" or englishClass =="MAGE" or englishClass == "PRIEST" then
			return true
		else
			return false
		end
	end
	
	return false
end
		  
function DrDamage:OnEnable()

	if DrDamage:Compatible( "Class" ) then
	
		--TEMP hack to fix bug in last commits
		if DrDamage.db.profile.DisplayType == "AvgDMG" then
			DrDamage.db.profile[DisplayType] = "Avg"
		end
	
		DrDamage:Hook( GameTooltip, "SetAction", true)
		DrDamage:SecureHook( GameTooltip, "SetSpell" )
		DrDamage:DmgUpdateTalents()
	else
		self:Print( "Your class is not supported by Dr. Damage." )
		do return end
	end
	
    	for i = 1, #changesDMG do
    		
    		if changesDMG[i] then
    			self:RegisterEvent( changesDMG[i], "DrOnEvent" )
    		end
  	end
	
	if DrDamage.db.profile.DisplayType ~= "Off" then
    		
    		DrDamage:UpdateAB()
    	
    		for i = 1, #changesAB do
    	
    			if changesAB[i] then
    				self:RegisterEvent( changesAB[i], "DrOnEvent" )
    			end
    		end
  	end   
end

function DrDamage:OnDisable()

	DrDamage:UpdateAB( true )
	talents = nil
	
end

function DrDamage:SetSpell( frame, slot )

	GT:SetSpell( slot, BOOKTYPE_SPELL )
	local spellName, spellRank = GetSpellName( slot, BOOKTYPE_SPELL );
	spellRank = string.sub( spellRank, 6 )
	
	if not spellRank then
		spellRank = "1"
	end
	
	DrDamage:DoTooltip( frame, spellName, spellRank )

end

function DrDamage:SetAction( frame, slot )

	self.hooks[frame].SetAction(frame, slot)

	local spellName = getglobal("GameTooltipTextLeft1"):GetText()
	local spellRank = getglobal("GameTooltipTextRight1"):GetText()
	
	if spellRank then
		spellRank = string.sub( spellRank, 6 )
	else
		spellRank = "1"
	end
	
	GT:SetAction( slot )
	DrDamage:DoTooltip( frame, spellName, spellRank )
end

function DrDamage:DoTooltip( frame, spellName, spellRank )

	if spellName and spellRank then
  	
  		local spellInfo = DrDamage:DetermineClass( "Spells" )
  		
		if not spellInfo then
			do return end
		end  		
  	
		if spellInfo[spellName] then
	
			local spell = spellInfo[spellName]
			local tableSpell = spell[spellRank]
			
			if not tableSpell then
				tableSpell = spell["1"]
			end
					
			if tableSpell then
			
				--NEEDS LOCALIZATION
				local _, _, ManaCost = GT:Find("(%d+) Mana", 2, 2)
				
				if ManaCost then
					ManaCost = tonumber( ManaCost )
				end
				
				local HealingSpell, MinDmg, MaxDmg, AvgDmg, AvgDmgCrit, CritRate, DotDmg, DPS, DPSC, SpellDam, SpellM, DamM, NextCrit, NextTenDam, IgniteDam = DrDamage:RawNumbers ( tableSpell, spellName )

				local DPM
				local DOOM
				local PlayerMana = UnitManaMax("player")
				
				if ManaCost and ManaCost > 0 then
					DPM = DrDamage:Round( AvgDmgCrit / ManaCost, 1 )
					
					if PlayerMana then
						DOOM = DrDamage:Round( DPM * PlayerMana, 1 )
					end
				end
					
				
				frame:AddLine(" ")
				frame:AddLine("Dr. Damage:", 1,1,1)
				frame:AddLine(" ")
				
				if HealingSpell then
					frame:AddDoubleLine("+Healing: ", SpellDam .. " (x" .. SpellM .. ")", 1, 1, 1, 0, 0.7, 0  )
				else
					frame:AddDoubleLine("+Dmg: ", SpellDam .. " (x" .. SpellM .. ")", 1, 1, 1, 0, 0.7, 0  )
				end
  				
  				if CritRate > 0 then
  					frame:AddDoubleLine("Crit:", CritRate .. "%", 1, 1, 1, 0, 0.7, 0 )
  				end	
  				
  				frame:AddDoubleLine("Avg (x".. DamM .. "):", AvgDmg .. " (".. MinDmg .."-".. MaxDmg ..")", 1, 1, 1, 0.8, 0.8, 0.9 )
  				
  				if AvgDmgCrit > AvgDmg then
  					frame:AddDoubleLine("Avg Crit/DoT:", AvgDmgCrit, 1, 1, 1, 0.8, 0.8, 0.9  )
  				end
  				
  				if IgniteDam > 0 then
  					frame:AddDoubleLine("Ignite:", IgniteDam, 1, 1, 1, 0.8, 0.8, 0.9 )
  				end
  				
   				if DotDmg > 0 then
  					frame:AddDoubleLine("DoT:", DotDmg, 1, 1, 1, 0.8, 0.8, 0.9 )
  				end 				
  				
  				if NextCrit > 0 then
  					if HealingSpell then
  						frame:AddDoubleLine("+1% Crit:", "+" .. NextCrit .. " (= " .. DrDamage:Round( ( NextCrit * 10 ) / NextTenDam, 1 ) .. " +healing)", 1, 1, 1, 0.3, 0.6, 0.5 )	
  					else
  						frame:AddDoubleLine("+1% Crit:", "+" .. NextCrit .. " (= " .. DrDamage:Round( ( NextCrit * 10 ) / NextTenDam, 1 ) .. " +dmg)", 1, 1, 1, 0.3, 0.6, 0.5 )	
  					end
  				end
  				
  				if HealingSpell then
  					frame:AddDoubleLine( "+10 Healing:", "+" .. NextTenDam, 1, 1, 1, 0.3, 0.6, 0.5 )
  				else
  					frame:AddDoubleLine( "+10 Dmg:", "+" .. NextTenDam, 1, 1, 1, 0.3, 0.6, 0.5 )
  				end
  				
  				
  				if DPSC ~= DPS then
  					if HealingSpell then
  						frame:AddDoubleLine("HPS/HPSC:", DPS .. "/" .. DPSC, 1, 1, 1, 0.8, 0.1, 0.1)
  					else
  						frame:AddDoubleLine("DPS/DPSC:", DPS .. "/" .. DPSC, 1, 1, 1, 0.8, 0.1, 0.1)
      					end
      				else
      					if HealingSpell then
      						frame:AddDoubleLine("HPS:", DPS, 1, 1, 1, 0.8, 0.1, 0.1)
      					else
      						frame:AddDoubleLine("DPS:", DPS, 1, 1, 1, 0.8, 0.1, 0.1)
      					end
      				end
      				
      				if DPM then
      					if HealingSpell then
      						frame:AddDoubleLine("HPM:", DPM, 1, 1, 1, 0.8, 0.1, 0.1 )
      					else
      						frame:AddDoubleLine("DPM:", DPM, 1, 1, 1, 0.8, 0.1, 0.1 )
      					end
      				end
      				
      				if DOOM then
      					if HealingSpell then
      						frame:AddDoubleLine("Healing until OOM:", DOOM, 1, 1, 1, 0.8, 0.1, 0.1 )
      					else
      						frame:AddDoubleLine("Damage until OOM:", DOOM, 1, 1, 1, 0.8, 0.1, 0.1 )
      					end
      				end
      				
      				frame:Show()			
			end  			
		end	  
	end
end


function dGet (n)
	return function () return DrDamage.db.profile[n] 
	end
end
function dSet (n, v)
	return function (v) DrDamage.db.profile[n] = v; DrDamage:UpdateAB(); 
	end
end

local DisplayTypes = { "Avg", "DPS", "DPSC", "Off" }

local options = { 
	type='group',
	args = {
		ActionBar = {
			type = "group", desc = "Options for the actionbar", name="ActionBar",
			args = {
				DisplayType = {
					name = "Display",
					type = "text",
					desc = "Choose what to display on the actionbar",
					validate = DisplayTypes,
					get = dGet("DisplayType"),
					set = dSet("DisplayType", value),
					order = 70,
				},
        			Update = {
            				name = "Update",
            				desc = "Forces update to actionbar",
            				type = "execute",
            				order = 75,
            				func = "UpdateAB",
    				}, 
				FontSize = {
					type = 'range',
					name = "Actionbar font size",
					min = 6,
					max = 14,
					step = 1,
					desc = "Set the font size",
					order = 71,
					get = dGet("FontSize"),
					set = dSet("FontSize", value),
				},
				FontPosition = {
					type = 'range',
					name = "Actionbar font position",
					min = -5,
					max = 5,
					step = 1,
					desc = "Set the font position",
					order = 72,
					get = dGet("FontPosition"),
					set = dSet("FontPosition", value),
				},    				
			},
		},
		Damage = {
			type = "group", desc = "Damage options", name="Damage",
			args = {
				ManualDmg = {
					type = 'toggle',
					name = "Manual variables",
					desc = "Allows you to manually set damage properties",
					order = 71,
					get = dGet( "ManualDmg" ),
					set = dSet( "ManualDmg", value ),
				},
				ManualPlus = {
					type = 'toggle',
					name = "Manual damage adds to detected",
					desc = "Allows you to set if manual damage should add to detected damage",
					order = 72,
					get = dGet( "ManualPlus" ),
					set = dSet( "ManualPlus", value ),
				}, 
				Healing = {
					type = 'range',
					name = "Healing",
					min = 0,
					max = 2000,
					step = 1,
					desc = "Input your +healing",
					order = 79,
					get = dGet("Healing"),
					set = dSet("Healing", value),
				},				
				SpellDamage = {
					type = 'range',
					name = "Spell Damage",
					min = 0,
					max = 2000,
					step = 1,
					desc = "Input your spell damage",
					order = 80,
					get = dGet("SpellDamage"),
					set = dSet("SpellDamage", value),
				},		
				FireDamage = {
					type = 'range',
					name = "Fire Damage",
					min = 0,
					max = 2000,
					step = 1,
					desc = "Input your fire +damage",
					order = 81,
					get = dGet("FireDamage"),
					set = dSet("FireDamage", value),		
		
				},
				ShadowDamage = {
					type = 'range',
					name = "Shadow Damage",
					min = 0,
					max = 2000,
					step = 1,
					desc = "Input your shadow +damage",
					order = 82,
					get = dGet("ShadowDamage"),
					set = dSet("ShadowDamage", value),		
				},
				FrostDamage = {
					type = 'range',
					name = "Frost Damage",
					min = 0,
					max = 2000,
					step = 1,
					desc = "Input your frost +damage",
					order = 83,
					get = dGet("FrostDamage"),
					set = dSet("FrostDamage", value),		
				},
				NatureDamage = {
					type = 'range',
					name = "Nature Damage",
					min = 0,
					max = 2000,
					step = 1,
					desc = "Input your nature +damage",
					order = 84,
					get = dGet("NatureDamage"),
					set = dSet("NatureDamage", value),		
				},
				ArcaneDamage = {
					type = 'range',
					name = "Arcane Damage",
					min = 0,
					max = 2000,
					step = 1,
					desc = "Input your arcane +damage",
					order = 85,
					get = dGet("ArcaneDamage"),
					set = dSet("ArcaneDamage", value),		
				},
				HolyDamage = {
					type = 'range',
					name = "Holy Damage",
					min = 0,
					max = 2000,
					step = 1,
					desc = "Input your holy +damage",
					order = 85,
					get = dGet("HolyDamage"),
					set = dSet("HolyDamage", value),		
				},					
				ItemCrit = {
					type = 'range',
					name = "Item Crit Bonus",
					min = 0,
					max = 1,
					isPercent = true,
					desc = "Input your crit bonus from items",
					order = 88,
					get = dGet("ItemCrit"),
					set = dSet("ItemCrit", value),					
				},
				ItemHit = {
					type = 'range',
					name = "Item Hit Bonus",
					min = 0,
					max = 1,
					isPercent = true,
					desc = "Input your hit bonus from items",
					order = 89,
					get = dGet("ItemHit"),
					set = dSet("ItemHit", value),	
				},
			},
		},
	},
};

DrDamage:RegisterChatCommand( { "/drdmg", "/drdamage" }, options);

local talents = {};

function DrDamage:DmgUpdateTalents()

	talents = {}
	local talentList = DrDamage:DetermineClass( "Talents" )
	
	if not talentList then
		do return end
	end
	
	local numTabs = GetNumTalentTabs();
	
	for t=1, numTabs do
	
    		local numTalents = GetNumTalents(t);
    		  		
	    	for i=1, numTalents do
	    	
	    		nameTalent, _, _, _, currRank, maxRank= GetTalentInfo(t,i);
	    		
	    	    	if currRank ~= 0 then
	    	    			    	    
	    	    		if talentList[nameTalent] then
	    	    		
	    	    			local v = talentList[nameTalent]		
	    	    			table.insert( talents, { Name = nameTalent, Rank = currRank } )
	    	    			--self:Print("Talent name: " .. nameTalent .. " Points: " .. currRank .. "/" .. maxRank )
	    	    		end    	    	
	    	    	end	    	    
	    	end
	end
end


function DrDamage:DetermineClass( requestType )

	localizedClass, englishClass = UnitClass("player");
	local ClassTable
	
	if englishClass == "WARLOCK" then
		ClassTable = WarlockTable
	elseif englishClass == "MAGE" then
		ClassTable = MageTable
	elseif englishClass == "PRIEST" then
		ClassTable = PriestTable
	else
		self:Print( "Unsupported class detected." )
	end
	
	if ClassTable then
	
		if requestType == "Talents" then
			return ClassTable["talentInfo"]	
		elseif requestType == "Spells" then
			return ClassTable["spellInfo"]
		end
	end
end
	

function DrDamage:UpdateAB( disable )

	local blizzAction
	local BT3 = false
	local Bongos = false
	local DAB = false
	local NAB = false
	local blizzAB = false

	if IsAddOnLoaded("Bartender3") then
		BT3 = true
	elseif IsAddOnLoaded( "Bongos" ) then
		Bongos = true
	elseif IsAddOnLoaded("DiscordActionBars") then
		DAB = true
	elseif IsAddOnLoaded("Nurfed_ActionBars") then
		NAB = true
	else
		blizzAB = true
		blizzAction = { "ActionButton", "BonusActionButton", "MultiBarBottomLeftButton", "MultiBarBottomRightButton","MultiBarRightButton", "MultiBarLeftButton", }
	end
	
	local spellInfo = DrDamage:DetermineClass( "Spells" )
	
	if not spellInfo then
		do return end
	end
	
	for i = 1, 120 do
	
		local ab
		local name
	
		if BT3 then
			name = "BT3Button"..i
			ab = Bartender3.actionbuttons
		elseif Bongos then
			name = "BActionButton"..i
		elseif DAB then
			name = "DAB_ActionButton_"..i
		elseif NAB then
			name = "Nurfed_ActionButton"..i
		elseif blizzAB then
			local abType = 1 + math.floor( ( i - 1 ) / 12 )
			local buttonNr = i - math.floor( ( i - 1 ) / 12 ) * 12
			
			if abType > #blizzAction then
				do return end
			end
			
			name = blizzAction[abType] .. buttonNr
		end
					
		local button = getglobal(name)
					
		if button then
			local children = { button:GetChildren() }
		
			if children then
   				for i = 1, #children do
   				
   					local frame = children[i]
   		
   					if frame and frame.text and frame:GetID() == 200 then
   						frame.text:Hide()
   					end
   					
   				end
   			end			
		end
			
		if button and button:IsVisible() and not disable then
		
			local id
		
			if BT3 and ab[i] and ab[i].state == "used" then
				id = ab[i].object:PagedID()
			elseif Bongos then
				id = button:GetPagedID()
			elseif DAB then
				id = button:GetActionID()
			elseif NAB then
				id = button:GetID()
			elseif blizzAB then
				id = ActionButton_GetPagedID( button )
			end
			
			if id then
			
				local gtype, pid, gsubtype = GetActionInfo(id)
			
				if gtype == "spell" then
			
					local spellName, spellRank = GetSpellName( pid, BOOKTYPE_SPELL );
			
					if spellInfo[spellName] then

						local spell = spellInfo[spellName]
						local tableSpell = spell[string.sub( spellRank, 6 )]
						
						if not tableSpell then
							tableSpell = spell["1"]
						end
		
						if tableSpell then

							HealingSpell, _, _, _, AvgDmgCrit, _, _, DPS, DPSC = DrDamage:RawNumbers ( tableSpell, spellName )
								
							if DrDamage.db.profile.DisplayType == "DPS" then
								DrDamage:CreateABtext( button, DPS, HealingSpell )
							elseif DrDamage.db.profile.DisplayType == "DPSC" then
								DrDamage:CreateABtext( button, DPSC, HealingSpell )
							elseif DrDamage.db.profile.DisplayType == "Avg" then
								DrDamage:CreateABtext( button, AvgDmgCrit, HealingSpell )
							end
						end				
					end			
				end
			end
		end
	end	
end


function DrDamage:CreateABtext( button, cDmg, HealingSpell )

	local children = { button:GetChildren() }
	local found = false
	
	if children then
   		for i = 1, #children do
   		
   			local frame = children[i]
   			
   			if frame and frame.text and frame:GetID() == 200 then
   				 				
   				frame.text:SetPoint("BOTTOMLEFT", frame, "BOTTOM", (DrDamage.db.profile.FontPosition - 15), 5)
   				frame.text:SetFont( STANDARD_TEXT_FONT, DrDamage.db.profile.FontSize, "OUTLINE" )
   				frame.text:SetText( cDmg )
   				frame.text:Show()
   				found = true
   			end
   		end
   	end
   		
   	if not found then
   		
		local dpsText = CreateFrame("Frame", "DrDamageText", button )
		dpsText:SetAllPoints( button )
		dpsText:SetFrameLevel(dpsText:GetFrameLevel() + 1)
		dpsText:SetID( 200 )
		dpsText.text = dpsText:CreateFontString( nil, "OVERLAY" )
		dpsText.text:SetFontObject(GameFontNormalSmall)
		dpsText.text:SetPoint("BOTTOMLEFT", dpsText, "BOTTOM", (DrDamage.db.profile.FontPosition - 15), 5)
		dpsText.text:SetFont(STANDARD_TEXT_FONT, DrDamage.db.profile.FontSize, "OUTLINE")
		
		if HealingSpell then
			dpsText.text:SetTextColor(0, 0.7, 0)
		else
			dpsText.text:SetTextColor(1, 1, 0.2)
		end
		
		dpsText.text:SetText( cDmg )
		dpsText.text:Show()
	end

end

function DrDamage:SpellCrit()

	if ManualDmg then
		local crit = GetSpellCritChanceFromIntellect() + DrDamage.db.profile.ItemCrit * 100
		return crit
	else
		local crit = GetSpellCritChanceFromIntellect() + GetCombatRatingBonus(11)
		return crit
	end
end

function DrDamage:GetPlusDamage( school )

	local spellDMG = GetSpellBonusDamage(1)
	
	if school == "Holy" then
  		return GetSpellBonusDamage(2)
	elseif school == "Fire" then
  		return GetSpellBonusDamage(3)
  	elseif school == "Nature" then
  		return GetSpellBonusDamage(4)
  	elseif school == "Frost" then
  		return GetSpellBonusDamage(5)
  	elseif school == "Shadow" then
  		return GetSpellBonusDamage(6)  		
   	elseif school == "Arcane" then
  		return GetSpellBonusDamage(7)
  	else
  		return spellDMG
  	end	
end

function DrDamage:RawNumbers( spell, spellName )

	if not spell or not spellName then
		do return end
	end
		
	local spellInfo = DrDamage:DetermineClass( "Spells" )
	
	if not spellInfo then
		do return end
	end	
	
	local baseSpell = (spellInfo[spellName])["0"]
	
	if not talents or #talents < 1 then
		DrDamage:DmgUpdateTalents()
	end
	
	local healingSpell = false
	
	if baseSpell.Healing then
		healingSpell = true
	end
	
	local spellDmg

	if healingSpell then
		spellDmg = GetSpellBonusHealing() 			-- +healing
	else
		spellDmg = DrDamage:GetPlusDamage( baseSpell.School ) 	--+all spelldamage
	end
	
		
	local critPerc = DrDamage:SpellCrit() 	--Calculate crit
	
	if DrDamage.db.profile.ManualDmg then
		
		if not DrDamage.db.profile.ManualPlus then
			
			if healingSpell then
				spellDmg = DrDamage.db.profile.Healing
			else
				spellDmg = DrDamage.db.profile.SpellDamage
			end
			
			critPerc = DrDamage.db.profile.ItemCrit
		else
			critPerc = critPerc + DrDamage.db.profile.ItemCrit
		end
		
		if not healingSpell then
		
			if ( baseSpell.School == "Shadow" ) then
				spellDmg = spellDmg + DrDamage.db.profile.ShadowDamage --add shadowdamage
			elseif ( baseSpell.School == "Fire" ) then
				spellDmg = spellDmg + DrDamage.db.profile.FireDamage --add firedamage
			elseif ( baseSpell.School == "Nature" ) then
				spellDmg = spellDmg + DrDamage.db.profile.NatureDamage --add naturedamage
			elseif ( baseSpell.School == "Frost" ) then
				spellDmg = spellDmg + DrDamage.db.profile.FrostDamage --add frostdamage
			elseif ( baseSpell.School == "Arcane" ) then
				spellDmg = spellDmg + DrDamage.db.profile.ArcaneDamage --add arcanedamage
			elseif ( baseSpell.School == "Holy" ) then
				spellDmg = spellDmg + DrDamage.db.profile.HolyDamage --add holydamage			
			end
		else
			spellDmg = spellDmg + DrDamage.db.profile.Healing
		end
	end
			
	local critM = 0.5			--Crit multiplier
	local bNukeDmg = 0			--Modifier for spells with a DOT portion
	local dmgM = 1				--Damage multiplier (total)
	local bDmgM = 1				--Base damage Modifier
	local spellDmgM				--Spell damage multiplier
	local igniteM = 0

	local castTime = 1.5
	
	if spell.castTime then
		castTime = spell.castTime
	elseif baseSpell.castTime then
		castTime = baseSpell.castTime
	end
	
	local eDuration = 1
	
	if spell.eDuration then
		eDuration = spell.eDuration
	elseif baseSpell.eDuration then
		eDuration = baseSpell.eDuration
	end
	
	if baseSpell.eDot then 
		spellDmgM = ( eDuration / 15)
	else
		spellDmgM = ( castTime / 3.5) 
	end
	
	--Modified factors?
	
	--Split between bonus to dot and nuke
	if baseSpell.hybridFactor then
		spellDmgM = spellDmgM * baseSpell.hybridFactor
	end
	
	--Factor that applies to all ranks of the spell
	if baseSpell.sFactor then
		spellDmgM = spellDmgM * baseSpell.sFactor
	end
	
	--Factor that applies to current rank of the spell -> Low ranks!
	if spell.sFactor then
		spellDmgM = spellDmgM * spell.sFactor
	end
	
	--Factor for lower rank spells
	if spell.spellLevel then
	
		local playerLevel = UnitLevel("player")
		local lowRankFactor = 1
	
		if healingSpell then
			lowRankFactor = ( spell.spellLevel + 6 ) / playerLevel
		else
			lowRankFactor = ( spell.spellLevel + 11.5 ) / playerLevel
		end
		
		if lowRankFactor < 1 then
			spellDmgM = spellDmgM * lowRankFactor
		end
	end
	
	local talentEffects = DrDamage:DetermineClass( "Talents" )
	
	if not talentEffects then
		do return end
	end
			
	for i, charTalent in ipairs( talents ) do
		
		local talentName = charTalent.Name
		local talentRank = charTalent.Rank
		
		if talentName then
					
			if talentEffects[talentName] then
				
				local effectTalent = talentEffects[talentName]
				
				function DoCalc( ModType, talentSpell, talentValue )
				
					if talentSpell == spellName or talentSpell == "All" then
					
						if ModType == "SpellDamage" then
							spellDmgM = spellDmgM + talentValue
						elseif ModType == "SpellDamageBonus" then
							spellDmgM = spellDmgM * ( talentValue + 1 )
						elseif ModType == "CastTime" then
							
							if baseSpell.CastMod then
								castTime = castTime - baseSpell.CastMod * talentValue
							else
								castTime = castTime - talentValue
							end
								
							if castTime < 1.5 then
								castTime = 1.5
							end
						elseif ModType == "BaseNuke" then
							bNukeDmg = bNukeDmg + talentValue
						elseif ModType == "Crit" then
							critPerc = critPerc + talentValue
						elseif ModType == "CritMultiplier" then
							critM = critM + talentValue
						elseif ModType == "Ignite" then
							igniteM = talentValue
						elseif ModType == "SpellDuration" then
							eDuration = eDuration + talentValue
						elseif not ModType then
							dmgM = dmgM + talentValue		
						end
					end				
				
				end
						
				for j, talentSpell in ipairs( effectTalent.Spells ) do
				
					local talentValue = talentRank * effectTalent.Effect
					
					DoCalc( effectTalent.ModType, talentSpell, talentValue  )
				end
				
				if effectTalent.Spells2 then
					for j, talentSpell in ipairs( effectTalent.Spells2 ) do
					
						local talentValue2
						
							if effectTalent.Effect2 then
								talentValue2 = talentRank * effectTalent.Effect2
							else
								talentValue2 = talentRank * effectTalent.Effect
							end
						
						DoCalc( effectTalent.ModType2, talentSpell, talentValue2  )
					end
				end	
			end
		end		
	end
	
	--Start calculating Buffs/Debuffs that affects damage not provided by blizzard API (What is missing?)
	
	if not healingSpell and DamageBuffs and DamageDebuffs then
	
		for buffName, index in SEA:DebuffIter("target") do
	
			if DamageDebuffs[buffName] then
			
				local buffData = DamageDebuffs[buffName]
			
				if buffData.School and buffData.School == baseSpell.School or buffData.School2 and buffData.School2 == baseSpell.School or not buffData.School then
					
					if buffData.Ranks then
						local _, _, _, rank = SEA:UnitHasDebuff( "target", buffName )
						local debuffRank = buffData.Ranks
				
							if rank then
								debuffRank = tonumber( string.sub( rank, 6 ) )
							end				
						
							if not buffData.ModType then
								dmgM = dmgM + debuffRank * buffData.Value
							end
					elseif buffData.Apps then
						local _, apps, _, _ = SEA:UnitHasDebuff( "target", buffName )
						
						if buffData.ModType == "Crit" then
							critPerc = critPerc + apps * buffData.Value
						elseif not buffData.ModType then
							dmgM = dmgM + apps * buffData.Value
						end
					elseif buffData.Apps2 then
						local _, apps, _, _ = SEA:UnitHasDebuff( "target", buffName )
						dmgM = dmgM + apps * buffData.Value2
					elseif not buffData.ModType then
						dmgM = dmgM + buffData.Value
					end	
					
					if buffData.Texture and buffData.ModType == "Texture" then
						local _, _, texture, _ = SEA:UnitHasDebuff( "target", buffName )
						if string.find( texture, buffData.Texture ) then
							dmgM = dmgM + buffData.Value
						end
					end

				end
			end
		end
		
		
		for buffName, index in SEA:BuffIter("player") do
		
			if DamageBuffs[buffName] then
			
				local buffData = DamageBuffs[buffName]
			
				if buffData.School and buffData.School == baseSpell.School or buffData.School2 and buffData.School2 == baseSpell.School or not buffData.School then
					
					if buffData.Ranks then
						local _, _, _, rank = SEA:UnitHasBuff( "player", buffName )
						local buffRank = buffData.Ranks
				
						if rank then
							buffRank = tonumber( string.sub( rank, 6 ) )
						end				
						
						if buffData.ModType == "MD" then
							
							if UnitCreatureFamily( "pet" ) then
								local petFamily = UnitCreatureFamily("pet")
					
								if petFamily == "Succubus" then
									dmgM = dmgM + 0.02 * buffRank
								elseif petFamily == "Felguard" then
									dmgM = dmgM + 0.01 * buffRank
								end
							end
						elseif not buffData.ModType then
							dmgM = dmgM + buffRank * buffData.Value
						end
					elseif buffData.Apps then
						local _, apps, _, _ = SEA:UnitHasBuff( "player", buffName )
						
						if not buffData.ModType then
							dmgM = dmgM + apps * buffData.Value
						end
					elseif buffData.ModType == "BaseDamage" then
						if buffData.Spell == spellName then
							bDmgM = bDmgM + buffData.Value
						end
					elseif buffData.ModType == "SpellDamage" then
						spellDmgM = spellDmgM + buffData.Value
					elseif buffData.ModType == "Boss" then
						dmgM = dmgM + buffData.Value
						
						if buffName == "Burning Adrenaline" then
							castTime = 1.5
						end
					elseif not buffData.ModType then
						dmgM = dmgM + buffData.Value
					end
				end
			end		
		end	
	end
	
	--Initialize variables that are calculated
	local calcAvgDmgCrit			--Calculated final damage with crit
	local calcAvgDmg			--Calculated final damage without crit
	local calcMinDmg			--Calculated dmg range: MIN
	local calcMaxDmg			--Calculated dmg range: MAX
	local calcDotDmg			--Calculated DOT portion (if applicable)
	local igniteDotDmg			--Calculated Ignite DOT Portion
	local avgIgnite				--Calculated Average Ignite
	
	
	function DmgCalc()
	
		calcAvgDmgCrit, calcAvgDmg, calcMinDmg, calcMaxDmg, calcDotDmg, igniteDotDmg, avgIgnite = 0, 0, 0, 0, 0, 0, 0
	
		if not baseSpell.sHits then
			calcMinDmg = dmgM * ( bDmgM * spell.bMin + ( spellDmg * spellDmgM ) )
			calcMaxDmg = dmgM * ( bDmgM * spell.bMax + ( spellDmg * spellDmgM ) )
		elseif not spell.hybridDotDmg then
			calcMinDmg = dmgM * ( bDmgM * spell.bMin * baseSpell.sHits + (  spellDmg * spellDmgM ) )
			calcMaxDmg = dmgM * ( bDmgM * spell.bMax * baseSpell.sHits + (  spellDmg * spellDmgM ) )
		else
			calcMinDmg = ( dmgM + bNukeDmg ) * ( bDmgM * spell.bMin + ( spellDmg * spellDmgM ) )
			calcMaxDmg = ( dmgM + bNukeDmg ) * ( bDmgM * spell.bMax + ( spellDmg * spellDmgM ) )
		end
	
		calcAvgDmg = (calcMinDmg + calcMaxDmg) / 2
		
		if baseSpell.eDot and baseSpell.eDuration and eDuration > baseSpell.eDuration then
			local bonusDotDamage = ( calcAvgDmg / baseSpell.eDuration ) * ( eDuration - baseSpell.eDuration )
			calcAvgDmg = calcAvgDmg + bonusDotDamage
		end
	
		if baseSpell.canCrit then
			calcAvgDmgCrit = calcAvgDmg + ( critPerc / 100 ) * calcAvgDmg * critM
		else
			calcAvgDmgCrit = calcAvgDmg
			critPerc = 0;
		end
	
		if spell.hybridDotDmg then
	
			local hybridFactor = 0
			local sFactor = 1
			
				if baseSpell.hybridFactor then
					hybridFactor = baseSpell.hybridFactor
				end
				
				if baseSpell.sFactor then
					sFactor = baseSpell.sFactor * sFactor
				end
				
				if spell.sFactor then
					sFactor = spell.sFactor * sFactor
				end
		
			calcDotDmg = dmgM * ( spell.hybridDotDmg + ( eDuration / 15 ) * ( 1 - hybridFactor ) * sFactor * spellDmg )
			spellDmgM = spellDmgM +  ( eDuration / 15 )
		end
		
		if baseSpell.Ignite then
			igniteDotDmg = igniteM * calcAvgDmgCrit
			avgIgnite = ( critPerc / 100 ) * igniteDotDmg
		end
		
		if spell.extraDotDmg then
		
			local xFactor = 0
			
			if baseSpell.extraDotF then
				xFactor = baseSpell.extraDotF
			end
			
			calcDotDmg = calcDotDmg + spell.extraDotDmg + xFactor * spellDmg --TODO: Verify factor
		end
		
		
		--Begin DPS calculations
		
		local calcDPS
		local calcDPT
	
		if baseSpell.eDot then
			calcDPS = calcAvgDmg / eDuration
		else
		
			if baseSpell.Ignite then
			
				calcDPS = calcAvgDmgCrit / castTime + avgIgnite / 4
				
				if spell.hybridDotDmg or spell.extraDotDmg then
					calcDPS = calcDPS + calcDotDmg / eDuration
				end
				
				calcAvgDmgCrit = calcAvgDmgCrit + avgIgnite + calcDotDmg
				
			elseif spell.hybridDotDmg or spell.extraDotDmg then
				calcDPS = calcAvgDmgCrit / castTime + calcDotDmg / eDuration
				calcAvgDmgCrit = calcAvgDmgCrit + calcDotDmg
			else
				calcDPS = calcAvgDmgCrit / castTime
			end
		end
		
		calcDPT = calcAvgDmgCrit / castTime
		
		return DrDamage:Round( calcMinDmg ), DrDamage:Round( calcMaxDmg ), DrDamage:Round( calcAvgDmg ), calcAvgDmgCrit, DrDamage:Round( critPerc, 2 ), DrDamage:Round( calcDotDmg ), DrDamage:Round( calcDPS ), DrDamage:Round( calcDPT ), DrDamage:Round( spellDmg ), DrDamage:Round( igniteDotDmg )
		
	end
	
	local returnMinDmg, returnMaxDmg, returnAvgDmg, returnAvgDmgCrit, returnCritPerc, returnDotDmg, returnDPS, returnDPTC, returnSpellDmg, returnIgniteDotDmg = DmgCalc()
	
	local returnNextCrit = 0
	local returnNextTenDam = 0
	
	if baseSpell.canCrit then
		critPerc = critPerc + 1
		local _, _, _, NextCrit = DmgCalc()
		returnNextCrit = DrDamage:Round( NextCrit - returnAvgDmgCrit, 1 )
	end
	
	if baseSpell.canCrit then
		critPerc = critPerc - 1
		spellDmg = spellDmg + 10
		local _,_,_, NextTen = DmgCalc()
		returnNextTenDam = DrDamage:Round( NextTen - returnAvgDmgCrit, 1 )
	else
		spellDmg = spellDmg + 10
		local _,_,_, NextTen = DmgCalc()
		returnNextTenDam = DrDamage:Round( NextTen - returnAvgDmgCrit, 1 )
	end
	
	
			
	return healingSpell, returnMinDmg, returnMaxDmg, returnAvgDmg, DrDamage:Round( returnAvgDmgCrit ), returnCritPerc, returnDotDmg, returnDPS, returnDPTC, returnSpellDmg, DrDamage:Round( spellDmgM, 2 ), DrDamage:Round( dmgM, 2), returnNextCrit, returnNextTenDam, returnIgniteDotDmg
end

function DrDamage:Round(x, y)

	local s
	
	if y then 
		x = x * math.pow(10, y) 
	end
	
	if (x > 0) then
		s = math.floor(x+0.5);
  	else
    		s = math.ceil(x-0.5);
  	end
  	
  	if y then 
  		s = s / math.pow(10, y) 
  	end
  	
  	return s
end

